//-----------------------------------------------------------------------------
// File name:       HWSoundDriver.cpp
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Included Files
//-----------------------------------------------------------------------------
#include "HW/HWSoundDriver.h"
#include "HW/HWWrapperFunctions.h"

#include "OS/OSDefines.h"
#include "OS/OSMutex.h"
#include "OS/OSTask.h"

#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <alsa/asoundlib.h>

//-----------------------------------------------------------------------------
// Constant Array and String Initialization
//-----------------------------------------------------------------------------
const std::string CHWSoundDriver::DEFAULT_HCTL_NAME = "hw:0";
const std::string CHWSoundDriver::PLAYBACK_VOLUME_CONTROL_NAME = "PCM";
const std::string CHWSoundDriver::CAPTURE_VOLUME_CONTROL_NAME = "PGA";
const std::string CHWSoundDriver::LEFT_PGA_MIXER_MIC3R_SWITCH_CONTROL_NAME = "Left PGA Mixer Mic3R";
const std::string CHWSoundDriver::RIGHT_PGA_MIXER_MIC3R_SWITCH_CONTROL_NAME = "Right PGA Mixer Mic3R";
const std::string CHWSoundDriver::CAPTURE_DATA_FILE_NAME = "/home/capture.dat";

const int16_t CHWSoundDriver::m_sineWave500Hz[CHWSoundDriver::SAMPLES_FOR_500_HZ_SINE_WAVE] =
{
    0, 960, 1916, 2863, 3799, 4718, 5617, 6492, 7340, 8155,
    8936, 9679, 10380, 11037, 11646, 12206, 12713, 13166, 13562, 13900,
    14179, 14397, 14554, 14648, 14680, 14648, 14554, 14397, 14179, 13901,
    13562, 13166, 12713, 12206, 11646, 11037, 10380, 9679, 8936, 8155,
    7340, 6492, 5617, 4718, 3799, 2863, 1916, 960, 0, -960,
    -1916, -2863, -3799, -4718, -5617, -6492, -7339, -8155, -8936, -9679,
    -10380, -11037, -11646, -12205, -12713, -13166, -13562, -13900, -14179, -14397,
    -14554, -14648, -14680, -14648, -14554, -14398, -14179, -13901, -13562, -13166,
    -12713, -12206, -11646, -11037, -10380, -9679, -8936, -8155, -7340, -6492,
    -5617, -4718, -3799, -2864, -1916, -960
};

const int16_t CHWSoundDriver::m_sineWave1000Hz[CHWSoundDriver::SAMPLES_FOR_1000_HZ_SINE_WAVE] =
{
    0, 1916, 3799, 5617, 7340, 8936, 10380, 11646, 12713, 13562,
    14179, 14554, 14680, 14554, 14179, 13562, 12713, 11646, 10380, 8936,
    7340, 5617, 3799, 1916, 0, -1916, -3799, -5617, -7339, -8936,
    -10380, -11646, -12713, -13562, -14179, -14554, -14680, -14554, -14179, -13562,
    -12713, -11646, -10380, -8936, -7340, -5617, -3799, -1916
};

const std::string CHWSoundDriver::ERROR_STRINGS[CHWSoundDriver::ERROR_CODE_MAX_VALUE] =
{
    "No Error",                             // ERROR_CODE_NO_ERROR
    "Unable to fill parms",                 // ERROR_CODE_UNABLE_TO_FILL_PARMS
    "Unable to set access type",            // ERROR_CODE_UNABLE_TO_SET_ACCESS_TYPE
    "Unable to set format",                 // ERROR_CODE_UNABLE_TO_SET_FORMAT_TYPE
    "Unable to set sample rate",            // ERROR_CODE_UNABLE_TO_SET_SAMPLING_RATE
    "Unable to set number of channels",     // ERROR_CODE_UNABLE_TO_SET_NUMBER_OF_CHANNELS
    "Unable to set buffer time",            // ERROR_CODE_UNABLE_TO_SET_BUFFER_TIME
    "Unable to set buffer size",            // ERROR_CODE_UNABLE_TO_GET_BUFFER_SIZE
    "Unable to set period time",            // ERROR_CODE_UNABLE_TO_SET_PERIOD_TIME
    "Unable to set period size",            // ERROR_CODE_UNABLE_TO_GET_PERIOD_SIZE
    "Unable to set parameters",             // ERROR_CODE_UNABLE_TO_SET_PARAMETERS
    "Unable to open playback device",       // ERROR_CODE_UNABLE_TO_OPEN_PLAYBACK_DEVICE
    "Unable to open capture device",        // ERROR_CODE_UNABLE_TO_OPEN_CAPTURE_DEVICE
    "Unable to open mixer device",          // ERROR_CODE_UNABLE_TO_OPEN_MIXER_DEVICE
    "Write error",                          // ERROR_CODE_WRITE_ERROR
    "Not all frames written",               // ERROR_CODE_NOT_ALL_FRAMES_WRITTEN
    "Read error",                           // ERROR_CODE_READ_ERROR
    "Not all frames read",                  // ERROR_CODE_NOT_ALL_FRAMES_READ
    "Too many poll descriptors",            // ERROR_CODE_TOO_MANY_POLL_DESCRIPTORS
    "Unable to get poll descriptors",       // ERROR_CODE_UNABLE_TO_GET_POLL_DESCRIPTORS
    "Unable to start capture",              // ERROR_CODE_UNABLE_TO_START_CAPTURE
    "Unable to open capture file",          // ERROR_CODE_UNABLE_TO_OPEN_CAPTURE_DATA_FILE
    "Unable to recover",                    // ERROR_CODE_UNABLE_TO_RECOVER
    "Underrun unknown state",               // ERROR_CODE_UNDERRUN_UNKNOWN_STATE
    "Unable to attach mixer",               // ERROR_CODE_UNABLE_TO_ATTACH_MIXER
    "Unable to register mixer",             // ERROR_CODE_UNABLE_TO_REGISTER_MIXER
    "Unable to load mixer",                 // ERROR_CODE_UNABLE_TO_LOAD_MIXER
    "Unable to find control element",       // ERROR_CODE_UNABLE_TO_FIND_CONTROL_ELEMENT
    "Unable to get control value",          // ERROR_CODE_UNABLE_TO_GET_CONTROL_VALUE
    "Unable to set control value",          // ERROR_CODE_UNABLE_TO_SET_CONTROL_VALUE
    "Unable to create capture file mutex",  // ERROR_CODE_UNABLE_TO_CREATE_CAPTURE_FILE_MUTEX
};

//-----------------------------------------------------------------------------
// Constructor/Destructor
//-----------------------------------------------------------------------------
CHWSoundDriver::CHWSoundDriver(void) :
    m_playbackThreadHandle(0),
    m_captureThreadHandle(0),
    m_playbackFileHandle(0),
    m_captureFileHandle(0),
    m_mixerFileHandle(0),
    m_playbackType(PLAYBACK_TYPE_500_HZ_TONE),
    m_captureType(CAPTURE_TYPE_NONE),
    m_playbackThreadActive(false),
    m_captureThreadActive(false),
    m_captureDataFileDescriptor(0),
    m_currentCaptureIndex(CIRCULAR_BUFFER_CAPTURE_STARTING_INDEX),
    m_currentPlaybackIndex(CIRCULAR_BUFFER_PLAYBACK_STARTING_INDEX),
    m_tone500HzPlaybackIndex(0)
{
    memset(m_captureCircularBuffer, 0, sizeof(m_captureCircularBuffer));
}

//-----------------------------------------------------------------------------
// Public Functions
//-----------------------------------------------------------------------------
uint8_t CHWSoundDriver::Initialize(void)
{
    int32_t result;
    uint8_t errorCode;
    bool mutexResult;

    result = snd_pcm_open(&m_playbackFileHandle, "default", SND_PCM_STREAM_PLAYBACK, 0);
    if (result == OS_SUCCESS)
    {
        errorCode = SetUpPCMDevice(m_playbackFileHandle);
    }
    else
    {
        errorCode = ERROR_CODE_UNABLE_TO_OPEN_PLAYBACK_DEVICE;
    }

    if (errorCode == ERROR_CODE_NO_ERROR)
    {
        result = snd_pcm_open(&m_captureFileHandle, "default", SND_PCM_STREAM_CAPTURE, 0);
        if (result == OS_SUCCESS)
        {
            errorCode = SetUpPCMDevice(m_captureFileHandle);
        }
        else
        {
            errorCode = ERROR_CODE_UNABLE_TO_OPEN_CAPTURE_DEVICE;
        }
    }

    if (errorCode == ERROR_CODE_NO_ERROR)
    {
        result = snd_mixer_open(&m_mixerFileHandle, 0);
        if (result == OS_SUCCESS)
        {
            errorCode = SetUpMixerDevice(m_mixerFileHandle);
        }
        else
        {
            errorCode = ERROR_CODE_UNABLE_TO_OPEN_MIXER_DEVICE;
        }
    }

    if (errorCode == ERROR_CODE_NO_ERROR)
    {
        mutexResult = COSMutex::CreateMutex(&m_captureFileMutex);
        if (!mutexResult)
        {
            errorCode = ERROR_CODE_UNABLE_TO_CREATE_CAPTURE_FILE_MUTEX;
        }
    }

    return errorCode;
}

//-----------------------------------------------------------------------------

void CHWSoundDriver::Cleanup(void)
{
    snd_pcm_drop(m_captureFileHandle);

    if (COSMutex::LockMutex(&m_captureFileMutex))
    {
        if (m_captureDataFileDescriptor > 0)
        {
            close(m_captureDataFileDescriptor);
        }
        COSMutex::DestroyMutex(&m_captureFileMutex);
    }

    snd_pcm_hw_free(m_playbackFileHandle);
    snd_pcm_hw_free(m_captureFileHandle);

    snd_pcm_close(m_playbackFileHandle);
    snd_pcm_close(m_captureFileHandle);

    snd_mixer_close(m_mixerFileHandle);
}

//-----------------------------------------------------------------------------

bool CHWSoundDriver::StartPlaybackTask(uint8_t playbackType)
{
    if (playbackType < PLAYBACK_TYPE_MAX_VALUE)
    {
        m_playbackType = playbackType;
    }
    else
    {
        m_playbackType = PLAYBACK_TYPE_500_HZ_TONE;
        m_tone500HzPlaybackIndex = 0;
    }

    if (!m_playbackThreadActive)
    {
        m_playbackThreadActive = COSTask::CreateTask((COSTask::OSTaskFunction)HWWRAP_PlaybackTaskFunctionWrapper,
                                                     COSTask::GetMinimumDefaultTaskStackSize(),
                                                     NULL,
                                                     THREAD_PRIORITY,
                                                     &m_playbackThreadHandle);
    }

    return m_playbackThreadActive;
}

//-----------------------------------------------------------------------------

bool CHWSoundDriver::StartCaptureTask(void)
{
    bool captureDataFileValid;

    m_captureType = CAPTURE_TYPE_FILE;

    // Open the file to write the captured data.
    captureDataFileValid = false;
    if (COSMutex::LockMutex(&m_captureFileMutex))
    {
        if (m_captureDataFileDescriptor <= 0)
        {
            m_captureDataFileDescriptor = open(CAPTURE_DATA_FILE_NAME.c_str(), (O_CREAT | O_WRONLY | O_TRUNC), (S_IRUSR | S_IWUSR));
            if (m_captureDataFileDescriptor != OS_FAILURE)
            {
                captureDataFileValid = true;
            }
        }
        COSMutex::UnlockMutex(&m_captureFileMutex);
    }

    // Start the capture thread if it is not already active.
    if (!m_captureThreadActive)
    {
        m_captureThreadActive = COSTask::CreateTask((COSTask::OSTaskFunction)HWWRAP_CaptureMicAudioTaskFunctionWrapper,
                                                    COSTask::GetMinimumDefaultTaskStackSize(),
                                                    NULL,
                                                    THREAD_PRIORITY,
                                                    &m_captureThreadHandle);
    }

    return (m_captureThreadActive && captureDataFileValid);
}

//-----------------------------------------------------------------------------

bool CHWSoundDriver::StopMicCapture(void)
{
    bool success;

    success = false;
    if (COSMutex::LockMutex(&m_captureFileMutex))
    {
        if (m_captureDataFileDescriptor > 0)
        {
            close(m_captureDataFileDescriptor);
            m_captureDataFileDescriptor = 0;
            success = true;
        }
        COSMutex::UnlockMutex(&m_captureFileMutex);
    }

    m_captureType = CAPTURE_TYPE_NONE;

    return success;
}

//-----------------------------------------------------------------------------

void CHWSoundDriver::StopSoundTask(void)
{
    m_playbackThreadActive = false;
    m_captureThreadActive = false;
}

//-----------------------------------------------------------------------------

bool CHWSoundDriver::StartLoopbackTest(void)
{
    m_currentCaptureIndex = CIRCULAR_BUFFER_CAPTURE_STARTING_INDEX;
    m_currentPlaybackIndex = CIRCULAR_BUFFER_PLAYBACK_STARTING_INDEX;
    memset(m_captureCircularBuffer, 0, sizeof(m_captureCircularBuffer));

    // Set the playback type to mic audio.
    m_playbackType = PLAYBACK_TYPE_MIC_AUDIO;
    m_captureType = CAPTURE_TYPE_BUFFER;

    // Start up both tasks if they are not already running.
    if (!m_playbackThreadActive)
    {
        m_playbackThreadActive = COSTask::CreateTask((COSTask::OSTaskFunction)HWWRAP_PlaybackTaskFunctionWrapper,
                                                     COSTask::GetMinimumDefaultTaskStackSize(),
                                                     NULL,
                                                     THREAD_PRIORITY,
                                                     &m_playbackThreadHandle);
    }

    if (!m_captureThreadActive)
    {
        m_captureThreadActive = COSTask::CreateTask((COSTask::OSTaskFunction)HWWRAP_CaptureMicAudioTaskFunctionWrapper,
                                                    COSTask::GetMinimumDefaultTaskStackSize(),
                                                    NULL,
                                                    THREAD_PRIORITY,
                                                    &m_captureThreadHandle);
    }

    return (m_playbackThreadActive && m_captureThreadActive);
}

//-----------------------------------------------------------------------------

bool CHWSoundDriver::StopLoopbackTest(void)
{
    // Set the playback type to silence.
    m_playbackType = PLAYBACK_TYPE_SILENCE;
    m_captureType = CAPTURE_TYPE_NONE;

    return true;
}

//-----------------------------------------------------------------------------

void CHWSoundDriver::PlaybackTaskFunction(void)
{
    struct pollfd playbackPollFD;
    snd_pcm_t* playbackHandle;
    snd_pcm_state_t state;
    int32_t playbackPollFDCount;
    int32_t result;
    uint8_t writeErrorCode;
    uint8_t taskErrorCode;
    bool underrunOccurred;
    bool recovered;
    bool taskError;

    taskError = false;
    playbackHandle = m_playbackFileHandle;
    taskErrorCode = ERROR_CODE_NO_ERROR;

    playbackPollFDCount = snd_pcm_poll_descriptors_count(playbackHandle);
    if (playbackPollFDCount != 1)
    {
        taskErrorCode = ERROR_CODE_TOO_MANY_POLL_DESCRIPTORS;
    }
    else
    {
        result = snd_pcm_poll_descriptors(playbackHandle, &playbackPollFD, playbackPollFDCount);
        if (result != playbackPollFDCount)
        {
            taskErrorCode = ERROR_CODE_UNABLE_TO_GET_POLL_DESCRIPTORS;
        }

        // Set the mixer controls necessary to play tone audio.
        if (taskErrorCode == ERROR_CODE_NO_ERROR)
        {
            taskErrorCode = SetPlaybackVolume(DEFAULT_PLAYBACK_VOLUME);
        }

        if (taskErrorCode == ERROR_CODE_NO_ERROR)
        {
            while (m_playbackThreadActive && !taskError)
            {
                writeErrorCode = GenerateOutput(playbackHandle);
                if (writeErrorCode == ERROR_CODE_WRITE_ERROR)
                {
                    // Attempt recovery
                    state = snd_pcm_state(playbackHandle);
                    recovered = AttemptToRecover(playbackHandle, state);
                    if (!recovered)
                    {
                        // A critical error has occurred.
                        taskError = true;
                        taskErrorCode = ERROR_CODE_UNABLE_TO_RECOVER;
                    }
                }

                underrunOccurred = CHWSoundDriver::WaitForPollEvent(playbackHandle, &playbackPollFD, playbackPollFDCount);
                if (underrunOccurred)
                {
                    state = snd_pcm_state(playbackHandle);
                    if ((state == SND_PCM_STATE_XRUN) || (state == SND_PCM_STATE_SUSPENDED))
                    {
                        printf("Recovery attempt: %d\r\n", state);
                        recovered = AttemptToRecover(playbackHandle, state);
                        if (!recovered)
                        {
                            // A critical error has occurred.
                            taskError = true;
                            taskErrorCode = ERROR_CODE_UNABLE_TO_RECOVER;
                        }
                        else
                        {
                            state = snd_pcm_state(playbackHandle);
                            printf("Recovered? %d\r\n", state);
                        }
                    }
                    else
                    {
                        // A critical error has occurred.
                        printf("Critical error\r\n");
                        taskError = true;
                        taskErrorCode = ERROR_CODE_UNDERRUN_UNKNOWN_STATE;
                    }
                }
            }
        }
    }
}

//-----------------------------------------------------------------------------

void CHWSoundDriver::CaptureTaskFunction(void)
{
    struct pollfd capturePollFD;
    snd_pcm_t* captureHandle;
    snd_pcm_state_t state;
    int32_t capturePollFDCount;
    int32_t result;
    uint8_t readErrorCode;
    uint8_t taskErrorCode;
    bool recovered;
    bool taskError;

    taskError = false;
    taskErrorCode = ERROR_CODE_NO_ERROR;
    captureHandle = m_captureFileHandle;

    capturePollFDCount = snd_pcm_poll_descriptors_count(captureHandle);
    if (capturePollFDCount != 1)
    {
        taskErrorCode = ERROR_CODE_TOO_MANY_POLL_DESCRIPTORS;
    }
    else
    {
        result = snd_pcm_poll_descriptors(captureHandle, &capturePollFD, capturePollFDCount);
        if (result != capturePollFDCount)
        {
            taskErrorCode = ERROR_CODE_UNABLE_TO_GET_POLL_DESCRIPTORS;
        }

        if (taskErrorCode == ERROR_CODE_NO_ERROR)
        {
            // Set the mixer controls necessary to capture mic audio.
            taskErrorCode = SetCaptureVolume(DEFAULT_CAPTURE_VOLUME);
            if (taskErrorCode == ERROR_CODE_NO_ERROR)
            {
                taskErrorCode = SetLeftPGAMixerMic3RSwitch(1);
            }
            if (taskErrorCode == ERROR_CODE_NO_ERROR)
            {
                taskErrorCode = SetLeftPGAMixerMic3RSwitch(1);
            }

            if (taskErrorCode == ERROR_CODE_NO_ERROR)
            {
                result = snd_pcm_start(captureHandle);
                if (result == OS_SUCCESS)
                {
                    while (m_captureThreadActive && !taskError)
                    {
                        readErrorCode = ReadInput(captureHandle);
                        if (readErrorCode == ERROR_CODE_READ_ERROR)
                        {
                            // Attempt recovery
                            state = snd_pcm_state(captureHandle);
                            recovered = AttemptToRecover(captureHandle, state);
                            if (!recovered)
                            {
                                // A critical error has occurred.
                                taskError = true;
                                taskErrorCode = ERROR_CODE_UNABLE_TO_RECOVER;
                            }
                        }
                    }
                }
                else
                {
                    taskErrorCode = ERROR_CODE_UNABLE_TO_START_CAPTURE;
                }
            }
        }
    }
}

//-----------------------------------------------------------------------------

std::string CHWSoundDriver::GetErrorCodeString(uint8_t errorEnum)
{
    const char* stringVal;

    if (errorEnum < ERROR_CODE_MAX_VALUE)
    {
        stringVal = ERROR_STRINGS[errorEnum].c_str();
    }
    else
    {
        stringVal = "Unknown error";
    }

    return stringVal;
}

//-----------------------------------------------------------------------------
// Private Functions
//-----------------------------------------------------------------------------
uint8_t CHWSoundDriver::SetUpPCMDevice(snd_pcm_t* handle)
{
    snd_pcm_hw_params_t* parms;
    int32_t result;
    uint8_t errorCode;

    errorCode = ERROR_CODE_NO_ERROR;

    snd_pcm_hw_params_alloca(&parms);
    result = snd_pcm_hw_params_any(handle, parms);
    if (result != OS_SUCCESS)
    {
        errorCode = ERROR_CODE_UNABLE_TO_FILL_PARMS;
    }

    if (errorCode == ERROR_CODE_NO_ERROR)
    {
        result = snd_pcm_hw_params_set_access(handle, parms, PCM_PARAMS_ACCESS_TYPE);
        if (result != OS_SUCCESS)
        {
            errorCode = ERROR_CODE_UNABLE_TO_SET_ACCESS_TYPE;
        }
    }

    if (errorCode == ERROR_CODE_NO_ERROR)
    {
        result = snd_pcm_hw_params_set_format(handle, parms, PCM_PARAMS_FORMAT_TYPE);
        if (result != OS_SUCCESS)
        {
            errorCode = ERROR_CODE_UNABLE_TO_SET_FORMAT_TYPE;
        }
    }

    if (errorCode == ERROR_CODE_NO_ERROR)
    {
        result = snd_pcm_hw_params_set_rate(handle, parms, PCM_PARAMS_SAMPLING_RATE, 0);
        if (result != OS_SUCCESS)
        {
            errorCode = ERROR_CODE_UNABLE_TO_SET_SAMPLING_RATE;
        }
    }

    if (errorCode == ERROR_CODE_NO_ERROR)
    {
        result = snd_pcm_hw_params_set_channels(handle, parms, PCM_PARAMS_NUMBER_OF_CHANNELS);
        if (result != OS_SUCCESS)
        {
            errorCode = ERROR_CODE_UNABLE_TO_SET_NUMBER_OF_CHANNELS;
        }
    }

    if (errorCode == ERROR_CODE_NO_ERROR)
    {
        result = snd_pcm_hw_params_set_period_size(handle, parms, PCM_PARAMS_PERIOD_SIZE_IN_FRAMES, 0);
        if (result != OS_SUCCESS)
        {
            errorCode = ERROR_CODE_UNABLE_TO_SET_PERIOD_TIME;
        }
    }

    if (errorCode == ERROR_CODE_NO_ERROR)
    {
        result = snd_pcm_hw_params_set_buffer_size(handle, parms, PCM_PARAMS_BUFFER_SIZE_IN_FRAMES);
        if (result != OS_SUCCESS)
        {
            errorCode = ERROR_CODE_UNABLE_TO_SET_BUFFER_TIME;
        }
    }

    if (errorCode == ERROR_CODE_NO_ERROR)
    {
        result = snd_pcm_hw_params(handle, parms);
        if (result != OS_SUCCESS)
        {
            errorCode = ERROR_CODE_UNABLE_TO_SET_PARAMETERS;
        }
    }

    return errorCode;
}

//-----------------------------------------------------------------------------

uint8_t CHWSoundDriver::SetUpMixerDevice(snd_mixer_t* handle)
{
    int32_t result;
    uint8_t errorCode;

    errorCode = ERROR_CODE_NO_ERROR;

    // Attach the mixer to the default sound card
    result = snd_mixer_attach(handle, DEFAULT_HCTL_NAME.c_str());
    if (result != OS_SUCCESS)
    {
        errorCode = ERROR_CODE_UNABLE_TO_ATTACH_MIXER;
    }

    if (errorCode == ERROR_CODE_NO_ERROR)
    {
        // Register the mixer
        result = snd_mixer_selem_register(handle, NULL, NULL);
        if (result != OS_SUCCESS)
        {
            errorCode = ERROR_CODE_UNABLE_TO_REGISTER_MIXER;
        }
    }

    if (errorCode == ERROR_CODE_NO_ERROR)
    {
        // Load the mixer
        result = snd_mixer_load(handle);
        if (result != OS_SUCCESS)
        {
            errorCode = ERROR_CODE_UNABLE_TO_LOAD_MIXER;
        }
    }

    return errorCode;
}

//-----------------------------------------------------------------------------

bool CHWSoundDriver::WaitForPollEvent(snd_pcm_t* handle, struct pollfd* fileDescriptor, uint32_t descriptorCount)
{
    uint16_t returnedEvents;
    bool underrunOccurred;
    bool waiting;

    underrunOccurred = false;
    waiting = true;

    while (waiting && m_playbackThreadActive)
    {
        // Wait forever for an event.
        poll(fileDescriptor, descriptorCount, POLL_TIMEOUT_VALUE_IN_MS);

        snd_pcm_poll_descriptors_revents(handle, fileDescriptor, descriptorCount, &returnedEvents);
        if (returnedEvents & POLLERR)
        {
            printf("Poll Error. E = %04X\r\n", returnedEvents);
            // We need to attempt recovery due to an underrun.
            underrunOccurred = true;
            waiting = false;
        }
        else if (returnedEvents & POLLOUT)
        {
            waiting = false;
        }
    }

    return underrunOccurred;
}

//-----------------------------------------------------------------------------

bool CHWSoundDriver::AttemptToRecover(snd_pcm_t* handle, snd_pcm_state_t currentState)
{
    int32_t result;
    bool recovered;

    recovered = false;

    if (currentState == SND_PCM_STATE_XRUN)
    {
        // An underrun has occurred.
        result = snd_pcm_prepare(handle);
        if (result == OS_FAILURE)
        {
            printf("Cannot recover from underrun\r\n");
        }
        else
        {
            recovered = true;
        }
    }
    else if (currentState == SND_PCM_STATE_SUSPENDED)
    {
        result = snd_pcm_resume(handle);
        while ((result == -EAGAIN) && m_playbackThreadActive)
        {
            // The resume call cannot complete (hardware is likely still suspended). Retry.
            sleep(1);
            result = snd_pcm_resume(handle);
        }

        if (result == OS_FAILURE)
        {
            result = snd_pcm_prepare(handle);
            if (result == OS_FAILURE)
            {
                printf("Cannot recover from suspend.\r\n");
            }
            else
            {
                recovered = true;
            }
        }
        else
        {
            recovered = true;
        }
    }

    return recovered;
}

//-----------------------------------------------------------------------------

uint8_t CHWSoundDriver::GenerateOutput(snd_pcm_t* handle)
{
    int16_t buffer[PCM_PARAMS_EXPECTED_PERIOD_SAMPLES];
    int32_t result;
    uint16_t toneIndex;
    uint16_t i;
    uint8_t errorCode;

    switch (m_playbackType)
    {
        case PLAYBACK_TYPE_500_HZ_TONE:
            for (i = 0; i < PCM_PARAMS_EXPECTED_PERIOD_SAMPLES; i++)
            {
                buffer[i] = m_sineWave500Hz[m_tone500HzPlaybackIndex];
                m_tone500HzPlaybackIndex = (m_tone500HzPlaybackIndex + 1) % SAMPLES_FOR_500_HZ_SINE_WAVE;

            }
            break;

        case PLAYBACK_TYPE_1000_HZ_TONE:
            toneIndex = 0;
            for (i = 0; i < PCM_PARAMS_EXPECTED_PERIOD_SAMPLES; i++)
            {
                buffer[i] = m_sineWave1000Hz[toneIndex];
                toneIndex = (toneIndex + 1) % SAMPLES_FOR_1000_HZ_SINE_WAVE;
            }
            break;

        case PLAYBACK_TYPE_MIC_AUDIO:
            memcpy(buffer, &(m_captureCircularBuffer[m_currentPlaybackIndex][0]), PCM_PARAMS_EXPECTED_PERIOD_SAMPLES * sizeof(int16_t));
            m_currentPlaybackIndex = (m_currentPlaybackIndex + 1) % CIRCULAR_BUFFER_NUM_BUFFERS;
            break;

        case PLAYBACK_TYPE_SILENCE:
        default:
            memset(buffer, 0, sizeof(buffer));
            break;
    }

    result = snd_pcm_writei(handle, buffer, PCM_PARAMS_EXPECTED_PERIOD_SAMPLES);
    if (result < 0)
    {
        // An error occurred.
        errorCode = ERROR_CODE_WRITE_ERROR;
        printf("snd_pcm_writei error %s\r\n", snd_strerror(result));
    }
    else if (result != PCM_PARAMS_EXPECTED_PERIOD_SAMPLES)
    {
        // Not all frames were written.
        errorCode = ERROR_CODE_NOT_ALL_FRAMES_WRITTEN;
        printf("snd_pcm_writei not all frames %d\r\n", result);
    }
    else
    {
        errorCode = ERROR_CODE_NO_ERROR;
    }

    return errorCode;
}

//-----------------------------------------------------------------------------

uint8_t CHWSoundDriver::ReadInput(snd_pcm_t* handle)
{
    int16_t buffer[PCM_PARAMS_EXPECTED_PERIOD_SAMPLES];
    int32_t result;
    uint8_t errorCode;

    result = snd_pcm_readi(handle, buffer, PCM_PARAMS_EXPECTED_PERIOD_SAMPLES);
    if (result < 0)
    {
        // An error occurred.
        errorCode = ERROR_CODE_READ_ERROR;
    }
    else if (result != PCM_PARAMS_EXPECTED_PERIOD_SAMPLES)
    {
        // Not all frames were read.
        errorCode = ERROR_CODE_NOT_ALL_FRAMES_READ;
    }
    else
    {
        errorCode = ERROR_CODE_NO_ERROR;
        switch (m_captureType)
        {
            case CAPTURE_TYPE_FILE:
                if (COSMutex::TryLockMutex(&m_captureFileMutex))
                {
                    if (m_captureDataFileDescriptor > 0)
                    {
                        write(m_captureDataFileDescriptor, buffer, sizeof(buffer));
                    }
                    COSMutex::UnlockMutex(&m_captureFileMutex);
                }
                break;

            case CAPTURE_TYPE_BUFFER:
                memcpy(&(m_captureCircularBuffer[m_currentCaptureIndex][0]), buffer, PCM_PARAMS_EXPECTED_PERIOD_SAMPLES * sizeof(int16_t));
                m_currentCaptureIndex = (m_currentCaptureIndex + 1) % CIRCULAR_BUFFER_NUM_BUFFERS;
                break;

            case CAPTURE_TYPE_NONE:
            default:
                break;
        }
    }

    return errorCode;
}

 //-----------------------------------------------------------------------------

uint8_t CHWSoundDriver::SetPlaybackVolume(uint8_t volume)
{
    snd_mixer_selem_id_t* id;
    snd_mixer_elem_t* element;
    int64_t value;
    int32_t result;
    uint8_t errorCode;

    errorCode = ERROR_CODE_NO_ERROR;

    // Allocate the ID space
    snd_mixer_selem_id_alloca(&id);

    // Set the identifier information and find the element
    snd_mixer_selem_id_set_index(id, 0);
    snd_mixer_selem_id_set_name(id, PLAYBACK_VOLUME_CONTROL_NAME.c_str());
    element = snd_mixer_find_selem(m_mixerFileHandle, id);
    if (element)
    {
        // The control element was found. Set the volume (must be a long value) for the front left and right channels.
        value = (int64_t)volume;
        result = snd_mixer_selem_set_playback_volume(element, SND_MIXER_SCHN_FRONT_LEFT, value);
        if (result != OS_SUCCESS)
        {
            errorCode = ERROR_CODE_UNABLE_TO_SET_CONTROL_VALUE;
        }
        else
        {
            result = snd_mixer_selem_set_playback_volume(element, SND_MIXER_SCHN_FRONT_RIGHT, value);
            if (result != OS_SUCCESS)
            {
                errorCode = ERROR_CODE_UNABLE_TO_SET_CONTROL_VALUE;
            }
        }
    }
    else
    {
        errorCode = ERROR_CODE_UNABLE_TO_FIND_CONTROL_ELEMENT;
    }

    return errorCode;
}

//-----------------------------------------------------------------------------

uint8_t CHWSoundDriver::GetPlaybackVolume(uint8_t* volume)
{
    snd_mixer_selem_id_t* id;
    snd_mixer_elem_t* element;
    int64_t value;
    int32_t result;
    uint8_t errorCode;

    errorCode = ERROR_CODE_NO_ERROR;

    // Allocate the ID space
    snd_mixer_selem_id_alloca(&id);

    // Set the identifier information and find the element
    snd_mixer_selem_id_set_index(id, 0);
    snd_mixer_selem_id_set_name(id, PLAYBACK_VOLUME_CONTROL_NAME.c_str());
    element = snd_mixer_find_selem(m_mixerFileHandle, id);
    if (element)
    {
        // The control element was found. Get the volume (must be a long value) from one of the channels
        // (assume it is the same on both left and right).
        result = snd_mixer_selem_get_playback_volume(element, SND_MIXER_SCHN_FRONT_LEFT, &value);
        if (result != OS_SUCCESS)
        {
            errorCode = ERROR_CODE_UNABLE_TO_GET_CONTROL_VALUE;
        }
        else
        {
            *volume = (uint8_t)value;
        }
    }
    else
    {
        errorCode = ERROR_CODE_UNABLE_TO_FIND_CONTROL_ELEMENT;
    }

    return errorCode;
}

//-----------------------------------------------------------------------------

uint8_t CHWSoundDriver::SetCaptureVolume(uint8_t volume)
{
    snd_mixer_selem_id_t* id;
    snd_mixer_elem_t* element;
    int64_t value;
    int32_t result;
    uint8_t errorCode;

    errorCode = ERROR_CODE_NO_ERROR;

    // Allocate the ID space
    snd_mixer_selem_id_alloca(&id);

    // Set the identifier information and find the element
    snd_mixer_selem_id_set_index(id, 0);
    snd_mixer_selem_id_set_name(id, CAPTURE_VOLUME_CONTROL_NAME.c_str());
    element = snd_mixer_find_selem(m_mixerFileHandle, id);
    if (element)
    {
        // The control element was found. Set the volume (must be a long value) for the front left and right channels.
        value = (int64_t)volume;
        result = snd_mixer_selem_set_capture_volume(element, SND_MIXER_SCHN_FRONT_LEFT, value);
        if (result != OS_SUCCESS)
        {
            errorCode = ERROR_CODE_UNABLE_TO_SET_CONTROL_VALUE;
        }
        else
        {
            result = snd_mixer_selem_set_capture_volume(element, SND_MIXER_SCHN_FRONT_RIGHT, value);
            if (result != OS_SUCCESS)
            {
                errorCode = ERROR_CODE_UNABLE_TO_SET_CONTROL_VALUE;
            }
        }
    }
    else
    {
        errorCode = ERROR_CODE_UNABLE_TO_FIND_CONTROL_ELEMENT;
    }

    return errorCode;
}

//-----------------------------------------------------------------------------

uint8_t CHWSoundDriver::GetCaptureVolume(uint8_t* volume)
{
    snd_mixer_selem_id_t* id;
    snd_mixer_elem_t* element;
    int64_t value;
    int32_t result;
    uint8_t errorCode;

    errorCode = ERROR_CODE_NO_ERROR;

    // Allocate the ID space
    snd_mixer_selem_id_alloca(&id);

    // Set the identifier information and find the element
    snd_mixer_selem_id_set_index(id, 0);
    snd_mixer_selem_id_set_name(id, CAPTURE_VOLUME_CONTROL_NAME.c_str());
    element = snd_mixer_find_selem(m_mixerFileHandle, id);
    if (element)
    {
        // The control element was found. Get the volume (must be a long value) from one of the channels
        // (assume it is the same on both left and right).
        result = snd_mixer_selem_get_capture_volume(element, SND_MIXER_SCHN_FRONT_LEFT, &value);
        if (result != OS_SUCCESS)
        {
            errorCode = ERROR_CODE_UNABLE_TO_GET_CONTROL_VALUE;
        }
        else
        {
            *volume = (uint8_t)value;
        }
    }
    else
    {
        errorCode = ERROR_CODE_UNABLE_TO_FIND_CONTROL_ELEMENT;
    }

    return errorCode;
}

//-----------------------------------------------------------------------------

uint8_t CHWSoundDriver::SetLeftPGAMixerMic3RSwitch(uint8_t switchValue)
{
    snd_mixer_selem_id_t* id;
    snd_mixer_elem_t* element;
    int32_t value;
    int32_t result;
    uint8_t errorCode;

    errorCode = ERROR_CODE_NO_ERROR;

    // Allocate the ID space
    snd_mixer_selem_id_alloca(&id);

    // Set the identifier information and find the element
    snd_mixer_selem_id_set_index(id, 0);
    snd_mixer_selem_id_set_name(id, LEFT_PGA_MIXER_MIC3R_SWITCH_CONTROL_NAME.c_str());
    element = snd_mixer_find_selem(m_mixerFileHandle, id);
    if (element)
    {
        // The control element was found. Set the switch for the mono channel.
        value = (int32_t)switchValue;
        result = snd_mixer_selem_set_playback_switch(element, SND_MIXER_SCHN_MONO, value);
        if (result != OS_SUCCESS)
        {
            errorCode = ERROR_CODE_UNABLE_TO_SET_CONTROL_VALUE;
        }
    }
    else
    {
        errorCode = ERROR_CODE_UNABLE_TO_FIND_CONTROL_ELEMENT;
    }

    return errorCode;
}

//-----------------------------------------------------------------------------

uint8_t CHWSoundDriver::GetLeftPGAMixerMic3RSwitch(uint8_t* switchValue)
{
    snd_mixer_selem_id_t* id;
    snd_mixer_elem_t* element;
    int32_t value;
    int32_t result;
    uint8_t errorCode;

    errorCode = ERROR_CODE_NO_ERROR;

    // Allocate the ID space
    snd_mixer_selem_id_alloca(&id);

    // Set the identifier information and find the element
    snd_mixer_selem_id_set_index(id, 0);
    snd_mixer_selem_id_set_name(id, LEFT_PGA_MIXER_MIC3R_SWITCH_CONTROL_NAME.c_str());
    element = snd_mixer_find_selem(m_mixerFileHandle, id);
    if (element)
    {
        // The control element was found. Get the switch value for the mono channel.
        result = snd_mixer_selem_get_playback_switch(element, SND_MIXER_SCHN_MONO, &value);
        if (result != OS_SUCCESS)
        {
            errorCode = ERROR_CODE_UNABLE_TO_GET_CONTROL_VALUE;
        }
        else
        {
            *switchValue = (uint8_t)value;
        }
    }
    else
    {
        errorCode = ERROR_CODE_UNABLE_TO_FIND_CONTROL_ELEMENT;
    }

    return errorCode;
}

//-----------------------------------------------------------------------------

uint8_t CHWSoundDriver::SetRightPGAMixerMic3RSwitch(uint8_t switchValue)
{
    snd_mixer_selem_id_t* id;
    snd_mixer_elem_t* element;
    int32_t value;
    int32_t result;
    uint8_t errorCode;

    errorCode = ERROR_CODE_NO_ERROR;

    // Allocate the ID space
    snd_mixer_selem_id_alloca(&id);

    // Set the identifier information and find the element
    snd_mixer_selem_id_set_index(id, 0);
    snd_mixer_selem_id_set_name(id, RIGHT_PGA_MIXER_MIC3R_SWITCH_CONTROL_NAME.c_str());
    element = snd_mixer_find_selem(m_mixerFileHandle, id);
    if (element)
    {
        // The control element was found. Set the switch for the mono channel.
        value = (int32_t)switchValue;
        result = snd_mixer_selem_set_playback_switch(element, SND_MIXER_SCHN_MONO, value);
        if (result != OS_SUCCESS)
        {
            errorCode = ERROR_CODE_UNABLE_TO_SET_CONTROL_VALUE;
        }
    }
    else
    {
        errorCode = ERROR_CODE_UNABLE_TO_FIND_CONTROL_ELEMENT;
    }

    return errorCode;
}

//-----------------------------------------------------------------------------

uint8_t CHWSoundDriver::GetRightPGAMixerMic3RSwitch(uint8_t* switchValue)
{
    snd_mixer_selem_id_t* id;
    snd_mixer_elem_t* element;
    int32_t value;
    int32_t result;
    uint8_t errorCode;

    errorCode = ERROR_CODE_NO_ERROR;

    // Allocate the ID space
    snd_mixer_selem_id_alloca(&id);

    // Set the identifier information and find the element
    snd_mixer_selem_id_set_index(id, 0);
    snd_mixer_selem_id_set_name(id, RIGHT_PGA_MIXER_MIC3R_SWITCH_CONTROL_NAME.c_str());
    element = snd_mixer_find_selem(m_mixerFileHandle, id);
    if (element)
    {
        // The control element was found. Get the switch value for the mono channel.
        result = snd_mixer_selem_get_playback_switch(element, SND_MIXER_SCHN_MONO, &value);
        if (result != OS_SUCCESS)
        {
            errorCode = ERROR_CODE_UNABLE_TO_GET_CONTROL_VALUE;
        }
        else
        {
            *switchValue = (uint8_t)value;
        }
    }
    else
    {
        errorCode = ERROR_CODE_UNABLE_TO_FIND_CONTROL_ELEMENT;
    }

    return errorCode;
}

//-----------------------------------------------------------------------------

void CHWSoundDriver::PrintControlElements(const char* filename)
{
    snd_mixer_elem_t* element;
    snd_mixer_selem_id_t* id;
    int32_t fileDescriptor;
    int32_t strLength;
    uint8_t i;
    bool writeToFile;
    char testString[150];

    // Write to a file if one is specified, otherwise print to the console.
    writeToFile = (strcmp(filename, "") != 0);
    if (writeToFile)
    {
        fileDescriptor = open(filename, O_CREAT | O_WRONLY, S_IRUSR | S_IWUSR);
        if (fileDescriptor == OS_FAILURE)
        {
            // The desired file couldn't be opened.
            writeToFile = false;
            printf("Error opening mixer control elements text file\r\n");
        }
    }

    // Allocate the ID space
    snd_mixer_selem_id_alloca(&id);

    // Loop through the mixer elements and print them in the desired format
    i = 0;
    for (element = snd_mixer_first_elem(m_mixerFileHandle); element; element = snd_mixer_elem_next(element))
    {
        if (writeToFile)
        {
            strLength = sprintf(testString, "Element %d name: %s\r\n", i, snd_mixer_selem_get_name(element));
            write(fileDescriptor, testString, strLength);
        }
        else
        {
            printf("Element %d name: %s\r\n", i, snd_mixer_selem_get_name(element));
        }
        i++;
    }

    if (writeToFile)
    {
        close(fileDescriptor);
    }
}
