//-----------------------------------------------------------------------------
// File name:       HWSoundDriver.h
//-----------------------------------------------------------------------------

#ifndef _HW_SOUND_DRIVER_H_
#define _HW_SOUND_DRIVER_H_

//-----------------------------------------------------------------------------
// Included Files
//-----------------------------------------------------------------------------
#include "OS/OSMutex.h"
#include "OS/OSTask.h"

#include <stdint.h>
#include <stdbool.h>
#include <string>
#include <alsa/asoundlib.h>

//-----------------------------------------------------------------------------
// Class Definition
//-----------------------------------------------------------------------------
class CHWSoundDriver
{
public:
    // Constants
    static const uint8_t SAMPLES_FOR_500_HZ_SINE_WAVE = 96;
    static const uint8_t SAMPLES_FOR_1000_HZ_SINE_WAVE = 48;

    // Enumerations
    enum ErrorCodes : uint8_t
    {
        ERROR_CODE_INITIAL_VALUE = 0,

        ERROR_CODE_NO_ERROR = ERROR_CODE_INITIAL_VALUE,
        ERROR_CODE_UNABLE_TO_FILL_PARMS,
        ERROR_CODE_UNABLE_TO_SET_ACCESS_TYPE,
        ERROR_CODE_UNABLE_TO_SET_FORMAT_TYPE,
        ERROR_CODE_UNABLE_TO_SET_SAMPLING_RATE,
        ERROR_CODE_UNABLE_TO_SET_NUMBER_OF_CHANNELS,
        ERROR_CODE_UNABLE_TO_SET_BUFFER_TIME,
        ERROR_CODE_UNABLE_TO_GET_BUFFER_SIZE,
        ERROR_CODE_UNABLE_TO_SET_PERIOD_TIME,
        ERROR_CODE_UNABLE_TO_GET_PERIOD_SIZE,
        ERROR_CODE_UNABLE_TO_SET_PARAMETERS,
        ERROR_CODE_UNABLE_TO_OPEN_PLAYBACK_DEVICE,
        ERROR_CODE_UNABLE_TO_OPEN_CAPTURE_DEVICE,
        ERROR_CODE_UNABLE_TO_OPEN_MIXER_DEVICE,
        ERROR_CODE_WRITE_ERROR,
        ERROR_CODE_NOT_ALL_FRAMES_WRITTEN,
        ERROR_CODE_READ_ERROR,
        ERROR_CODE_NOT_ALL_FRAMES_READ,
        ERROR_CODE_TOO_MANY_POLL_DESCRIPTORS,
        ERROR_CODE_UNABLE_TO_GET_POLL_DESCRIPTORS,
        ERROR_CODE_UNABLE_TO_START_CAPTURE,
        ERROR_CODE_UNABLE_TO_OPEN_CAPTURE_DATA_FILE,
        ERROR_CODE_UNABLE_TO_RECOVER,
        ERROR_CODE_UNDERRUN_UNKNOWN_STATE,
        ERROR_CODE_UNABLE_TO_ATTACH_MIXER,
        ERROR_CODE_UNABLE_TO_REGISTER_MIXER,
        ERROR_CODE_UNABLE_TO_LOAD_MIXER,
        ERROR_CODE_UNABLE_TO_FIND_CONTROL_ELEMENT,
        ERROR_CODE_UNABLE_TO_GET_CONTROL_VALUE,
        ERROR_CODE_UNABLE_TO_SET_CONTROL_VALUE,
        ERROR_CODE_UNABLE_TO_CREATE_CAPTURE_FILE_MUTEX,

        ERROR_CODE_MAX_VALUE
    };

    enum PlaybackType : uint8_t
    {
        PLAYBACK_TYPE_INITIAL_VALUE = 0,

        PLAYBACK_TYPE_SILENCE = PLAYBACK_TYPE_INITIAL_VALUE,
        PLAYBACK_TYPE_500_HZ_TONE,
        PLAYBACK_TYPE_1000_HZ_TONE,
        PLAYBACK_TYPE_MIC_AUDIO,

        PLAYBACK_TYPE_MAX_VALUE
    };

    enum CaptureType : uint8_t
    {
        CAPTURE_TYPE_INITIAL_VALUE = 0,

        CAPTURE_TYPE_NONE = CAPTURE_TYPE_INITIAL_VALUE,
        CAPTURE_TYPE_FILE,
        CAPTURE_TYPE_BUFFER,

        CAPTURE_TYPE_MAX_VALUE
    };

    // Constructor/Destructor
    CHWSoundDriver(void);
    ~CHWSoundDriver(void) = default;

    // Functions
    uint8_t Initialize(void);
    void Cleanup(void);

    bool StartPlaybackTask(uint8_t playbackType);
    bool StartCaptureTask(void);
    bool StopMicCapture(void);
    void StopSoundTask(void);
    bool StartLoopbackTest(void);
    bool StopLoopbackTest(void);

    void PlaybackTaskFunction(void);
    void CaptureTaskFunction(void);

    std::string GetErrorCodeString(uint8_t errorEnum);

private:
    // Constants
    static const uint8_t THREAD_PRIORITY = 80;

    static const snd_pcm_access_t PCM_PARAMS_ACCESS_TYPE = SND_PCM_ACCESS_RW_INTERLEAVED;
    static const snd_pcm_format_t PCM_PARAMS_FORMAT_TYPE = SND_PCM_FORMAT_S16_LE;
    static const uint32_t PCM_PARAMS_SAMPLING_RATE = 48000;
    static const uint8_t PCM_PARAMS_NUMBER_OF_CHANNELS = 1;
    static const uint32_t PCM_PARAMS_PERIOD_TIME_MS = 5;                                                // Period time in ms (length of time between HW interrupts)
    static const uint32_t PCM_PARAMS_PERIOD_SIZE_IN_FRAMES = ((PCM_PARAMS_SAMPLING_RATE / 1000) * PCM_PARAMS_PERIOD_TIME_MS);
    static const uint32_t PCM_PARAMS_BUFFER_SIZE_IN_FRAMES = (PCM_PARAMS_PERIOD_SIZE_IN_FRAMES * 2);    // Buffer size should be 2 * period size
    static const uint32_t PCM_PARAMS_EXPECTED_PERIOD_SAMPLES = (PCM_PARAMS_PERIOD_SIZE_IN_FRAMES * PCM_PARAMS_NUMBER_OF_CHANNELS);

    static const uint8_t DEFAULT_PLAYBACK_VOLUME = 127;                                                 // Range is 0 to 127
    static const uint8_t DEFAULT_CAPTURE_VOLUME = 90;                                                   // Range is 0 to 127.

    static const int32_t POLL_TIMEOUT_VALUE_IN_MS = (PCM_PARAMS_PERIOD_TIME_MS * 2);                    // Twice as long as expected HW interrupt.

    static const uint8_t CIRCULAR_BUFFER_NUM_BUFFERS = 3;
    static const uint8_t CIRCULAR_BUFFER_CAPTURE_STARTING_INDEX = 0;
    static const uint8_t CIRCULAR_BUFFER_PLAYBACK_STARTING_INDEX = 2;

    static const std::string DEFAULT_HCTL_NAME;
    static const std::string PLAYBACK_VOLUME_CONTROL_NAME;
    static const std::string CAPTURE_VOLUME_CONTROL_NAME;
    static const std::string LEFT_PGA_MIXER_MIC3R_SWITCH_CONTROL_NAME;
    static const std::string RIGHT_PGA_MIXER_MIC3R_SWITCH_CONTROL_NAME;
    static const std::string CAPTURE_DATA_FILE_NAME;

    static const int16_t m_sineWave500Hz[SAMPLES_FOR_500_HZ_SINE_WAVE];
    static const int16_t m_sineWave1000Hz[SAMPLES_FOR_1000_HZ_SINE_WAVE];

    static const std::string ERROR_STRINGS[ERROR_CODE_MAX_VALUE];

    // Functions
    uint8_t SetUpPCMDevice(snd_pcm_t* handle);
    uint8_t SetUpMixerDevice(snd_mixer_t* handle);

    bool WaitForPollEvent(snd_pcm_t* handle, struct pollfd* fileDescriptor, uint32_t descriptorCount);
    bool AttemptToRecover(snd_pcm_t* handle, snd_pcm_state_t currentState);
    uint8_t GenerateOutput(snd_pcm_t* handle);
    uint8_t ReadInput(snd_pcm_t* handle);

    // Mixer Functions
    uint8_t SetPlaybackVolume(uint8_t volume);
    uint8_t GetPlaybackVolume(uint8_t* volume);
    uint8_t SetCaptureVolume(uint8_t volume);
    uint8_t GetCaptureVolume(uint8_t* volume);
    uint8_t SetLeftPGAMixerMic3RSwitch(uint8_t switchValue);
    uint8_t GetLeftPGAMixerMic3RSwitch(uint8_t* switchValue);
    uint8_t SetRightPGAMixerMic3RSwitch(uint8_t switchValue);
    uint8_t GetRightPGAMixerMic3RSwitch(uint8_t* switchValue);

    // Debug Functions
    void PrintControlElements(const char* filename);

    // Variables
    COSTask::OSTaskHandle m_playbackThreadHandle;
    COSTask::OSTaskHandle m_captureThreadHandle;
    COSMutex::OSMutex m_captureFileMutex;
    snd_pcm_t* m_playbackFileHandle;
    snd_pcm_t* m_captureFileHandle;
    snd_mixer_t* m_mixerFileHandle;
    uint8_t m_playbackType;
    uint8_t m_captureType;
    bool m_playbackThreadActive;
    bool m_captureThreadActive;
    int32_t m_captureDataFileDescriptor;

    int16_t m_captureCircularBuffer[CIRCULAR_BUFFER_NUM_BUFFERS][PCM_PARAMS_EXPECTED_PERIOD_SAMPLES];
    uint8_t m_currentCaptureIndex;
    uint8_t m_currentPlaybackIndex;

    uint8_t m_tone500HzPlaybackIndex;
};

#endif // _HW_SOUND_DRIVER_H_
