//-----------------------------------------------------------------------------
// File name:       OSTask.cpp
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Included Files
//-----------------------------------------------------------------------------
#include "OS/OSDefines.h"
#include "OS/OSTask.h"

#include <stdint.h>
#include <stdbool.h>

#include "pthread.h"
#include "sched.h"
#include "time.h"
#include "limits.h"
#include <string.h>
#include <errno.h>

//-----------------------------------------------------------------------------
// Public Static Functions
//-----------------------------------------------------------------------------
bool COSTask::CreateTask(OSTaskFunction taskFunctionPointer,
                         uint32_t taskStackSize,
                         void* taskParameters,
                         uint8_t taskPriority,
                         OSTaskHandle* createdTaskHandle)
{
    bool taskCreated = false;
    pthread_attr_t threadAttributesRecord;
    struct sched_param schedulerParms;
    int32_t result;

    result = pthread_attr_init(&threadAttributesRecord);
    if (result == OS_SUCCESS)
    {
        if (taskStackSize < PTHREAD_STACK_MIN)
        {
            taskStackSize = PTHREAD_STACK_MIN;
        }

        // Set the stack size of the thread.
        result = pthread_attr_setstacksize(&threadAttributesRecord, taskStackSize);
        if (result == OS_SUCCESS)
        {
            // Set the scheduler policy to first in, first out
            result = pthread_attr_setschedpolicy(&threadAttributesRecord, SCHED_FIFO);
            if (result == OS_SUCCESS)
            {
                // Set the priority level for this thread.
                schedulerParms.sched_priority = taskPriority;
                result = pthread_attr_setschedparam(&threadAttributesRecord, &schedulerParms);
                if (result == OS_SUCCESS)
                {
                    // Use the scheduler parameters from the attribute instead of inheriting from the parent.
                    result = pthread_attr_setinheritsched(&threadAttributesRecord, PTHREAD_EXPLICIT_SCHED);
                }
            }
        }
    }

    if (result == OS_SUCCESS)
    {
        // If setting the stack size was successful, then create the thread.
        result = pthread_create(createdTaskHandle,
                                &threadAttributesRecord,
                                taskFunctionPointer,
                                taskParameters);
    }

    if (result == OS_SUCCESS)
    {
        taskCreated = true;
    }

    return taskCreated;
}

//-----------------------------------------------------------------------------

void COSTask::DestroyTask(OSTaskHandle taskHandle)
{
    if (taskHandle != 0)
    {
        // Only cancel a thread if it has a valid taskHandle. Otherwise, let the task function
        // complete, thus cleaning up the thread.
        pthread_cancel(taskHandle);
    }
}

//-----------------------------------------------------------------------------

void COSTask::TaskSleep(uint32_t sleepTimeInMilliseconds)
{
    struct timespec sleepTimeSpec;
    struct timespec remainingIntervalSpec;
    int16_t result;

    memset(&remainingIntervalSpec, 0, sizeof(remainingIntervalSpec));
    sleepTimeSpec.tv_sec = (sleepTimeInMilliseconds / 1000);
    sleepTimeSpec.tv_nsec = ((sleepTimeInMilliseconds % 1000) * 1000000);

    result = nanosleep(&sleepTimeSpec, &remainingIntervalSpec);

    while ((result == OS_FAILURE) && (errno == EINTR))
    {
        // The thread's pause was interrupted by a signal. Resume the sleep.
        memcpy(&sleepTimeSpec, &remainingIntervalSpec, sizeof(remainingIntervalSpec));
        memset(&remainingIntervalSpec, 0, sizeof(remainingIntervalSpec));
        result = nanosleep(&sleepTimeSpec, &remainingIntervalSpec);
    }
}

//-----------------------------------------------------------------------------

void COSTask::TaskYield(void)
{
    sched_yield();
}

//-----------------------------------------------------------------------------

uint32_t COSTask::GetMinimumTaskStackSize(void)
{
    uint32_t minStackSize;

    // The stack size must be greater than or equal to PTHREAD_STACK_MIN.
    minStackSize = PTHREAD_STACK_MIN;

    return minStackSize;
}

//-----------------------------------------------------------------------------

uint32_t COSTask::GetMinimumDefaultTaskStackSize(void)
{
    uint32_t stackSize;

    stackSize = COSTask::GetMinimumTaskStackSize();
    if (stackSize == 0)
    {
        stackSize = DEFAULT_MINIMUM_TASK_STACK_SIZE;
    }

    return stackSize;
}

//-----------------------------------------------------------------------------

uint32_t COSTask::ValidateTaskStackSize(uint32_t requestedSize)
{
    uint32_t stackSize;

    stackSize = COSTask::GetMinimumTaskStackSize();
    if (stackSize < requestedSize)
    {
        // If the request is greater than the minimum required, then use the requested size.
        stackSize = requestedSize;
    }

    return stackSize;
}
